@@ -19,6 +19,11 @@
 package org.apache.zookeeper.server.quorum;
 
 import static org.apache.zookeeper.test.ClientBase.CONNECTION_TIMEOUT;
+import static org.junit.Assert.assertEquals;
+import static org.mockito.Matchers.any;
+import static org.mockito.Mockito.doCallRealMethod;
+import static org.mockito.Mockito.mock;
+import static org.mockito.Mockito.when;
 
 import java.io.BufferedReader;
 import java.io.BufferedWriter;
@@ -32,6 +37,7 @@ import java.io.IOException;
 import java.net.InetSocketAddress;
 import java.nio.ByteBuffer;
 import java.nio.channels.SocketChannel;
+import java.util.HashMap;
 import java.util.Map;
 import java.util.concurrent.TimeUnit;
 import java.util.regex.Pattern;
@@ -39,15 +45,19 @@ import java.util.regex.Pattern;
 import org.apache.log4j.Layout;
 import org.apache.log4j.Level;
 import org.apache.log4j.Logger;
+import org.apache.log4j.PatternLayout;
 import org.apache.log4j.WriterAppender;
 import org.apache.zookeeper.CreateMode;
 import org.apache.zookeeper.KeeperException;
 import org.apache.zookeeper.PortAssignment;
+import org.apache.zookeeper.Watcher;
 import org.apache.zookeeper.ZooDefs.OpCode;
 import org.apache.zookeeper.ZooKeeper;
 import org.apache.zookeeper.ZooDefs.Ids;
 import org.apache.zookeeper.ZooKeeper.States;
+import org.apache.zookeeper.common.Time;
 import org.apache.zookeeper.common.AtomicFileOutputStream;
+import org.apache.zookeeper.server.persistence.FileTxnSnapLog;
 import org.apache.zookeeper.server.quorum.Leader.Proposal;
 import org.apache.zookeeper.server.util.ZxidUtils;
 import org.apache.zookeeper.test.ClientBase;
@@ -114,7 +124,7 @@ public class QuorumPeerMainTest extends QuorumPeerTestBase {
         waitForOne(zk, States.CONNECTED);
         zk.create("/foo_q1", "foobar1".getBytes(), Ids.OPEN_ACL_UNSAFE,
                 CreateMode.PERSISTENT);
-        Assert.assertEquals(new String(zk.getData("/foo_q1", null, null)), "foobar1");
+        assertEquals(new String(zk.getData("/foo_q1", null, null)), "foobar1");
         zk.close();
 
         zk = new ZooKeeper("127.0.0.1:" + CLIENT_PORT_QP2,
@@ -122,7 +132,7 @@ public class QuorumPeerMainTest extends QuorumPeerTestBase {
         waitForOne(zk, States.CONNECTED);
         zk.create("/foo_q2", "foobar2".getBytes(), Ids.OPEN_ACL_UNSAFE,
                 CreateMode.PERSISTENT);
-        Assert.assertEquals(new String(zk.getData("/foo_q2", null, null)), "foobar2");
+        assertEquals(new String(zk.getData("/foo_q2", null, null)), "foobar2");
         zk.close();
 
         q1.shutdown();
@@ -245,15 +255,8 @@ public class QuorumPeerMainTest extends QuorumPeerTestBase {
         numServers = 3;
         servers = LaunchServers(numServers);
         String path = "/hzxidtest";
-        int leader=-1;
+        int leader = servers.findLeader();
 
-        // find the leader
-        for (int i=0; i < numServers; i++) {
-            if (servers.mt[i].main.quorumPeer.leader != null) {
-                leader = i;
-            }
-        }
-        
         // make sure there is a leader
         Assert.assertTrue("There should be a leader", leader >=0);
 
@@ -306,7 +309,7 @@ public class QuorumPeerMainTest extends QuorumPeerTestBase {
         // validate that the old value is there and not the new one
         output = servers.zk[nonleader].getData(path+leader, false, null);
 
-        Assert.assertEquals(
+        assertEquals(
                 "Expecting old value 1 since 2 isn't committed yet",
                 output[0], 1);
 
@@ -322,81 +325,217 @@ public class QuorumPeerMainTest extends QuorumPeerTestBase {
 
         // make sure it doesn't have the new value that it alone had logged
         output = servers.zk[leader].getData(path+leader, false, null);
-        Assert.assertEquals(
+        assertEquals(
                 "Validating that the deposed leader has rolled back that change it had written",
                 output[0], 1);
         
         // make sure the leader has the subsequent changes that were made while it was offline
         output = servers.zk[leader].getData(path+nonleader, false, null);
-        Assert.assertEquals(
+        assertEquals(
                 "Validating that the deposed leader caught up on changes it missed",
                 output[0], 2);
     }
 
-    private void waitForOne(ZooKeeper zk, States state) throws InterruptedException {
-    	while(zk.getState() != state) {
-    		Thread.sleep(500);
-    	}
+    /**
+     * This test validates that if a quorum member determines that it is leader without the support of the rest of the
+     * quorum (the other members do not believe it to be the leader) it will stop attempting to lead and become a follower.
+     *
+     * @throws IOException
+     * @throws InterruptedException
+     */
+    @Test
+    public void testElectionFraud() throws IOException, InterruptedException {
+        // capture QuorumPeer logging
+        ByteArrayOutputStream os = new ByteArrayOutputStream();
+        String loggingPattern = ((PatternLayout) Logger.getRootLogger().getAppender("CONSOLE").getLayout()).getConversionPattern();
+        WriterAppender appender = new WriterAppender(new PatternLayout(loggingPattern), os);
+        appender.setThreshold(Level.INFO);
+        Logger qlogger = Logger.getLogger(QuorumPeer.class);
+        qlogger.addAppender(appender);
+
+        numServers = 3;
+
+        // used for assertions later
+        boolean foundLeading = false;
+        boolean foundLooking = false;
+        boolean foundFollowing = false;
+
+        try {
+          // spin up a quorum, we use a small ticktime to make the test run faster
+          servers = LaunchServers(numServers, 500);
+
+          // find the leader
+          int trueLeader = servers.findLeader();
+          Assert.assertTrue("There should be a leader", trueLeader >= 0);
+
+          // find a follower
+          int falseLeader = (trueLeader + 1) % numServers;
+          Assert.assertTrue("All servers should join the quorum", servers.mt[falseLeader].main.quorumPeer.follower != null);
+
+          // to keep the quorum peer running and force it to go into the looking state, we kill leader election
+          // and close the connection to the leader
+          servers.mt[falseLeader].main.quorumPeer.electionAlg.shutdown();
+          servers.mt[falseLeader].main.quorumPeer.follower.getSocket().close();
+
+          // wait for the falseLeader to disconnect
+          waitForOne(servers.zk[falseLeader], States.CONNECTING);
+
+          // convince falseLeader that it is the leader
+          servers.mt[falseLeader].main.quorumPeer.setPeerState(QuorumPeer.ServerState.LEADING);
+
+          // provide time for the falseleader to realize no followers have connected
+          // (this is twice the timeout used in Leader#getEpochToPropose)
+          Thread.sleep(2 * servers.mt[falseLeader].main.quorumPeer.initLimit * servers.mt[falseLeader].main.quorumPeer.tickTime);
+
+          // Restart leader election
+          servers.mt[falseLeader].main.quorumPeer.startLeaderElection();
+
+          // The previous client connection to falseLeader likely closed, create a new one
+          servers.zk[falseLeader] = new ZooKeeper("127.0.0.1:" + servers.mt[falseLeader].getClientPort(), ClientBase.CONNECTION_TIMEOUT, this);
+
+          // Wait for falseLeader to rejoin the quorum
+          waitForOne(servers.zk[falseLeader], States.CONNECTED);
+
+          // and ensure trueLeader is still the leader
+          Assert.assertTrue(servers.mt[trueLeader].main.quorumPeer.leader != null);
+
+          // Look through the logs for output that indicates the falseLeader is LEADING, then LOOKING, then FOLLOWING
+          LineNumberReader r = new LineNumberReader(new StringReader(os.toString()));
+          Pattern leading = Pattern.compile(".*myid=" + falseLeader + ".*LEADING.*");
+          Pattern looking = Pattern.compile(".*myid=" + falseLeader + ".*LOOKING.*");
+          Pattern following = Pattern.compile(".*myid=" + falseLeader + ".*FOLLOWING.*");
+
+          String line;
+          while ((line = r.readLine()) != null) {
+            if (!foundLeading) {
+              foundLeading = leading.matcher(line).matches();
+            } else if(!foundLooking) {
+              foundLooking = looking.matcher(line).matches();
+            } else if (following.matcher(line).matches()){
+              foundFollowing = true;
+              break;
+            }
+          }
+        } finally {
+          qlogger.removeAppender(appender);
+        }
+
+        Assert.assertTrue("falseLeader never attempts to become leader", foundLeading);
+        Assert.assertTrue("falseLeader never gives up on leadership", foundLooking);
+        Assert.assertTrue("falseLeader never rejoins the quorum", foundFollowing);
     }
-    
-	private void waitForAll(ZooKeeper[] zks, States state) throws InterruptedException {
-		int iterations = 10;
-		boolean someoneNotConnected = true;
-        while(someoneNotConnected) {
-        	if (iterations-- == 0) {
-        		ClientBase.logAllStackTraces();
-			throw new RuntimeException("Waiting too long");
-        	}
-        	
-        	someoneNotConnected = false;
-        	for(ZooKeeper zk: zks) {
-        		if (zk.getState() != state) {
-        			someoneNotConnected = true;
-        		}
-        	}
-        	Thread.sleep(1000);
+
+    private void waitForOne(ZooKeeper zk, States state) throws InterruptedException {
+        int iterations = ClientBase.CONNECTION_TIMEOUT / 500;
+        while (zk.getState() != state) {
+            if (iterations-- == 0) {
+                throw new RuntimeException("Waiting too long " + zk.getState() + " != " + state);
+            }
+            Thread.sleep(500);
+        }
+    }
+
+    private void waitForAll(Servers servers, States state) throws InterruptedException {
+        waitForAll(servers.zk, state);
+    }
+
+    private void waitForAll(ZooKeeper[] zks, States state) throws InterruptedException {
+        int iterations = ClientBase.CONNECTION_TIMEOUT / 1000;
+        boolean someoneNotConnected = true;
+        while (someoneNotConnected) {
+            if (iterations-- == 0) {
+                ClientBase.logAllStackTraces();
+                throw new RuntimeException("Waiting too long");
+            }
+
+            someoneNotConnected = false;
+            for (ZooKeeper zk : zks) {
+                if (zk.getState() != state) {
+                    someoneNotConnected = true;
+                    break;
+                }
+            }
+            Thread.sleep(1000);
         }
 	}
 
     // This class holds the servers and clients for those servers
-  	private class Servers {
-  	    MainThread mt[];
-  	    ZooKeeper zk[];
-  	}
-  	
-  	/**
-  	 * This is a helper function for launching a set of servers
-  	 *  
-  	 * @param numServers
-  	 * @return
-  	 * @throws IOException
-  	 * @throws InterruptedException
-  	 */
-  	private Servers LaunchServers(int numServers) throws IOException, InterruptedException {
-  	    int SERVER_COUNT = numServers;
-  	    Servers svrs = new Servers();
-  	    final int clientPorts[] = new int[SERVER_COUNT];
-  	    StringBuilder sb = new StringBuilder();
-  	    for(int i = 0; i < SERVER_COUNT; i++) {
-  	        clientPorts[i] = PortAssignment.unique();
-  	        sb.append("server."+i+"=127.0.0.1:"+PortAssignment.unique()+":"+PortAssignment.unique()+"\n");
-  	    }
-  	    String quorumCfgSection = sb.toString();
-  
-  	    MainThread mt[] = new MainThread[SERVER_COUNT];
-  	    ZooKeeper zk[] = new ZooKeeper[SERVER_COUNT];
-  	    for(int i = 0; i < SERVER_COUNT; i++) {
-  	        mt[i] = new MainThread(i, clientPorts[i], quorumCfgSection);
-  	        mt[i].start();
-  	        zk[i] = new ZooKeeper("127.0.0.1:" + clientPorts[i], ClientBase.CONNECTION_TIMEOUT, this);
-  	    }
-  
-  	    waitForAll(zk, States.CONNECTED);
-  
-  	    svrs.mt = mt;
-  	    svrs.zk = zk;
-  	    return svrs;
-  	}
+    private class Servers {
+        MainThread mt[];
+        ZooKeeper zk[];
+        int[] clientPorts;
+
+        public void shutDownAllServers() throws InterruptedException {
+            for (MainThread t: mt) {
+                t.shutdown();
+            }
+        }
+
+        public void restartAllServersAndClients(Watcher watcher) throws IOException {
+            for (MainThread t : mt) {
+                if (!t.isAlive()) {
+                    t.start();
+                }
+            }
+            for (int i = 0; i < zk.length; i++) {
+                restartClient(i, watcher);
+            }
+        }
+
+        public void restartClient(int clientIndex, Watcher watcher) throws IOException {
+            zk[clientIndex] = new ZooKeeper("127.0.0.1:" + clientPorts[clientIndex], ClientBase.CONNECTION_TIMEOUT, watcher);
+        }
+
+        public int findLeader() {
+            for (int i = 0; i < mt.length; i++) {
+                if (mt[i].main.quorumPeer.leader != null) {
+                    return i;
+                }
+            }
+            return -1;
+        }
+    }
+
+    private Servers LaunchServers(int numServers) throws IOException, InterruptedException {
+  	    return LaunchServers(numServers, null);
+    }
+
+    /**
+     * This is a helper function for launching a set of servers
+     *
+     * @param numServers the number of servers
+     * @param tickTime A ticktime to pass to MainThread
+     * @return
+     * @throws IOException
+     * @throws InterruptedException
+     */
+    private Servers LaunchServers(int numServers, Integer tickTime) throws IOException, InterruptedException {
+        int SERVER_COUNT = numServers;
+        Servers svrs = new Servers();
+        svrs.clientPorts = new int[SERVER_COUNT];
+        StringBuilder sb = new StringBuilder();
+        for(int i = 0; i < SERVER_COUNT; i++) {
+            svrs.clientPorts[i] = PortAssignment.unique();
+            sb.append("server."+i+"=127.0.0.1:"+PortAssignment.unique()+":"+PortAssignment.unique()+"\n");
+        }
+        String quorumCfgSection = sb.toString();
+
+        svrs.mt = new MainThread[SERVER_COUNT];
+        svrs.zk = new ZooKeeper[SERVER_COUNT];
+        for(int i = 0; i < SERVER_COUNT; i++) {
+            if (tickTime != null) {
+                svrs.mt[i] = new MainThread(i, svrs.clientPorts[i], quorumCfgSection, new HashMap<String, String>(), tickTime);
+            } else {
+                svrs.mt[i] = new MainThread(i, svrs.clientPorts[i], quorumCfgSection);
+            }
+            svrs.mt[i].start();
+            svrs.restartClient(i, this);
+        }
+
+        waitForAll(svrs, States.CONNECTED);
+
+        return svrs;
+    }
 
 
     /**
@@ -458,6 +597,135 @@ public class QuorumPeerMainTest extends QuorumPeerTestBase {
         Assert.assertTrue("complains about host", found);
     }
 
+    @Test
+    public void testValidIpv6AddressInQuorum() throws Exception {
+        assumeIPv6Available();
+
+        ClientBase.setupTestEnv();
+
+        // setup the logger to capture all logs
+        Layout layout =
+                Logger.getRootLogger().getAppender("CONSOLE").getLayout();
+        ByteArrayOutputStream os = new ByteArrayOutputStream();
+        WriterAppender appender = new WriterAppender(layout, os);
+        appender.setImmediateFlush(true);
+        appender.setThreshold(Level.INFO);
+        Logger qlogger = Logger.getLogger("org.apache.zookeeper.server.quorum");
+        qlogger.addAppender(appender);
+
+        try {
+            final int CLIENT_PORT_QP1 = PortAssignment.unique();
+            final int CLIENT_PORT_QP2 = PortAssignment.unique();
+
+            String quorumCfgSection =
+                    "server.1=127.0.0.1:" + PortAssignment.unique()
+                    + ":" + PortAssignment.unique()
+                    + "\nserver.2=[0:0:0:0:0:0:0:1]:" + PortAssignment.unique()
+                    + ":" + PortAssignment.unique();
+
+            MainThread q1 = new MainThread(1, CLIENT_PORT_QP1, quorumCfgSection);
+            MainThread q2 = new MainThread(2, CLIENT_PORT_QP2, quorumCfgSection);
+
+            q1.start();
+            q2.start();
+
+            Assert.assertTrue("waiting for server 1 being up",
+                    ClientBase.waitForServerUp("127.0.0.1:" + CLIENT_PORT_QP1,
+                            ClientBase.CONNECTION_TIMEOUT));
+
+            Assert.assertTrue("waiting for server 2 being up",
+                    ClientBase.waitForServerUp("[0:0:0:0:0:0:0:1]:" + CLIENT_PORT_QP1,
+                            ClientBase.CONNECTION_TIMEOUT));
+
+            q1.shutdown();
+            q2.shutdown();
+
+            Assert.assertTrue("waiting for server 1 down",
+                    ClientBase.waitForServerDown("127.0.0.1:" + CLIENT_PORT_QP1,
+                            ClientBase.CONNECTION_TIMEOUT));
+
+            Assert.assertTrue("waiting for server 2 down",
+                    ClientBase.waitForServerDown("[0:0:0:0:0:0:0:1]:" + CLIENT_PORT_QP1,
+                            ClientBase.CONNECTION_TIMEOUT));
+
+        } finally {
+            qlogger.removeAppender(appender);
+        }
+
+        os.close();
+        LineNumberReader r = new LineNumberReader(new StringReader(os.toString()));
+        String line;
+        boolean found = false;
+        Pattern p =
+                Pattern.compile(".*Resolved hostname: 0:0:0:0:0:0:0:1.*");
+        while ((line = r.readLine()) != null) {
+            found = p.matcher(line).matches();
+            if (found) {
+                break;
+            }
+        }
+        Assert.assertTrue("IPv6 address resolved", found);
+    }
+
+    @Test
+    public void testInvalidIpv6AddressInQuorum() throws Exception {
+        assumeIPv6Available();
+
+        ClientBase.setupTestEnv();
+
+        // setup the logger to capture all logs
+        Layout layout =
+                Logger.getRootLogger().getAppender("CONSOLE").getLayout();
+        ByteArrayOutputStream os = new ByteArrayOutputStream();
+        WriterAppender appender = new WriterAppender(layout, os);
+        appender.setImmediateFlush(true);
+        appender.setThreshold(Level.INFO);
+        Logger qlogger = Logger.getLogger("org.apache.zookeeper.server.quorum");
+        qlogger.addAppender(appender);
+
+        try {
+            final int CLIENT_PORT_QP1 = PortAssignment.unique();
+
+            String quorumCfgSection =
+                    "server.1=127.0.0.1:" + PortAssignment.unique()
+                    + ":" + PortAssignment.unique()
+                    + "\nserver.2=[0:0:0:0:0:0:0:1:" + PortAssignment.unique()
+                    + ":" + PortAssignment.unique();
+
+            MainThread q1 = new MainThread(1, CLIENT_PORT_QP1, quorumCfgSection);
+            q1.start();
+
+            boolean isup =
+                    ClientBase.waitForServerUp("127.0.0.1:" + CLIENT_PORT_QP1,
+                            30000);
+
+            Assert.assertFalse("Server never came up", isup);
+
+            q1.shutdown();
+
+            Assert.assertTrue("waiting for server 1 down",
+                    ClientBase.waitForServerDown("127.0.0.1:" + CLIENT_PORT_QP1,
+                            ClientBase.CONNECTION_TIMEOUT));
+
+        } finally {
+            qlogger.removeAppender(appender);
+        }
+
+        os.close();
+        LineNumberReader r = new LineNumberReader(new StringReader(os.toString()));
+        String line;
+        boolean found = false;
+        Pattern p =
+                Pattern.compile(".*QuorumPeerConfig\\$ConfigException.*");
+        while ((line = r.readLine()) != null) {
+            found = p.matcher(line).matches();
+            if (found) {
+                break;
+            }
+        }
+        Assert.assertTrue("complains about configuration", found);
+    }
+
     /**
      * Verify handling of inconsistent peer type
      */
@@ -474,7 +742,7 @@ public class QuorumPeerMainTest extends QuorumPeerTestBase {
         Logger qlogger = Logger.getLogger("org.apache.zookeeper.server.quorum");
         qlogger.addAppender(appender);
 
-        // test the most likely situation only: server is stated as observer in 
+        // test the most likely situation only: server is stated as observer in
         // servers list, but there's no "peerType=observer" token in config
         try {
             final int CLIENT_PORT_QP1 = PortAssignment.unique();
@@ -532,7 +800,7 @@ public class QuorumPeerMainTest extends QuorumPeerTestBase {
             Pattern.compile(".*Peer type from servers list.* doesn't match peerType.*");
         Pattern pObserve = Pattern.compile(".*OBSERVING.*");
         while ((line = r.readLine()) != null) {
-            if (pWarn.matcher(line).matches()) { 
+            if (pWarn.matcher(line).matches()) {
                 warningPresent = true;
             }
             if (pObserve.matcher(line).matches()) {
@@ -542,12 +810,12 @@ public class QuorumPeerMainTest extends QuorumPeerTestBase {
                 break;
             }
         }
-        Assert.assertTrue("Should warn about inconsistent peer type", 
+        Assert.assertTrue("Should warn about inconsistent peer type",
                 warningPresent && defaultedToObserver);
     }
 
     /**
-     * verify if bad packets are being handled properly 
+     * verify if bad packets are being handled properly
      * at the quorum port
      * @throws Exception
      */
@@ -594,7 +862,7 @@ public class QuorumPeerMainTest extends QuorumPeerTestBase {
         waitForOne(zk, States.CONNECTED);
         zk.create("/foo_q1", "foobar1".getBytes(), Ids.OPEN_ACL_UNSAFE,
                 CreateMode.PERSISTENT);
-        Assert.assertEquals(new String(zk.getData("/foo_q1", null, null)), "foobar1");
+        assertEquals(new String(zk.getData("/foo_q1", null, null)), "foobar1");
         zk.close();
         q1.shutdown();
         q2.shutdown();
@@ -685,9 +953,9 @@ public class QuorumPeerMainTest extends QuorumPeerTestBase {
         q1.start();
         // Let the notifications timeout
         Thread.sleep(30000);
-        long start = System.currentTimeMillis();
+        long start = Time.currentElapsedTime();
         q1.shutdown();
-        long end = System.currentTimeMillis();
+        long end = Time.currentElapsedTime();
         if ((end - start) > maxwait) {
            Assert.fail("QuorumPeer took " + (end -start) +
                     " to shutdown, expected " + maxwait);
@@ -749,7 +1017,7 @@ public class QuorumPeerMainTest extends QuorumPeerTestBase {
                 new File(servers.mt[i].dataDir, "version-2"),
                 QuorumPeer.CURRENT_EPOCH_FILENAME);
             LOG.info("Validating current epoch: " + servers.mt[i].dataDir);
-            Assert.assertEquals("Current epoch should be 11.", 11,
+            assertEquals("Current epoch should be 11.", 11,
                                 readLongFromFile(currentEpochFile));
         }
 
@@ -795,4 +1063,192 @@ public class QuorumPeerMainTest extends QuorumPeerTestBase {
         Assert.assertFalse("updatingEpoch file should get deleted",
                            updatingEpochFile.exists());
     }
+
+    @Test
+    public void testNewFollowerRestartAfterNewEpoch() throws Exception {
+        numServers = 3;
+
+        servers = LaunchServers(numServers);
+        waitForAll(servers.zk, States.CONNECTED);
+        String inputString = "test";
+        byte[] input = inputString.getBytes();
+        byte[] output;
+        String path = "/newepochzxidtest";
+
+        // Create a couple of nodes
+        servers.zk[0].create(path, input, Ids.OPEN_ACL_UNSAFE, CreateMode.PERSISTENT);
+        servers.zk[0].setData(path, input, -1);
+
+        // make sure the updates indeed committed. If it is not
+        // the following statement will throw.
+        output = servers.zk[1].getData(path, false, null);
+
+        // Shutdown every one
+        for (int i=0; i < numServers; i++) {
+            servers.mt[i].shutdown();
+        }
+
+        LOG.info("resetting follower");
+        MainThread follower = servers.mt[0];
+        // delete followers information
+        File followerDataDir = new File(follower.dataDir, "version-2");
+        for(File file: followerDataDir.listFiles()) {
+            LOG.info("deleting " + file.getName());
+            file.delete();
+        }
+
+        // Startup everyone except follower, wait for election.
+        for (int i=1; i < numServers; i++) {
+            servers.mt[i].start();
+        }
+        for (int i=1; i < numServers; i++) {
+            waitForOne(servers.zk[i], States.CONNECTED);
+        }
+
+        follower.start();
+        waitForAll(servers.zk, States.CONNECTED); // snapshot should be recieved
+
+        follower.shutdown();
+        follower.start();
+
+        Assert.assertFalse(follower.mainFailed.await(10, TimeUnit.SECONDS));
+        waitForAll(servers.zk, States.CONNECTED);
+    }
+
+    @Test
+    public void testDataDirAndDataLogDir() throws Exception {
+        // Arrange
+        QuorumPeerConfig configMock = mock(QuorumPeerConfig.class);
+        when(configMock.getDataDir()).thenReturn("/tmp/zookeeper");
+        when(configMock.getDataLogDir()).thenReturn("/tmp/zookeeperLog");
+
+        QuorumPeer qpMock = mock(QuorumPeer.class);
+
+        doCallRealMethod().when(qpMock).setTxnFactory(any(FileTxnSnapLog.class));
+        when(qpMock.getTxnFactory()).thenCallRealMethod();
+        InjectableQuorumPeerMain qpMain = new InjectableQuorumPeerMain(qpMock);
+
+        // Act
+        qpMain.runFromConfig(configMock);
+
+        // Assert
+        FileTxnSnapLog txnFactory = qpMain.getQuorumPeer().getTxnFactory();
+        assertEquals("/tmp/zookeeperLog/version-2", txnFactory.getDataDir().getAbsolutePath());
+        assertEquals("/tmp/zookeeper/version-2", txnFactory.getSnapDir().getAbsolutePath());
+    }
+
+    private class InjectableQuorumPeerMain extends QuorumPeerMain {
+        QuorumPeer qp;
+
+        InjectableQuorumPeerMain(QuorumPeer qp) {
+            this.qp = qp;
+        }
+
+        @Override
+        protected QuorumPeer getQuorumPeer() {
+            return qp;
+        }
+    }
+
+    @Test
+    public void testFailedTxnAsPartOfQuorumLoss() throws Exception {
+        final int LEADER_TIMEOUT_MS = 10000;
+        // 1. start up server and wait for leader election to finish
+        ClientBase.setupTestEnv();
+        final int SERVER_COUNT = 3;
+        servers = LaunchServers(SERVER_COUNT);
+
+        waitForAll(servers, States.CONNECTED);
+
+        // we need to shutdown and start back up to make sure that the create session isn't the first transaction since
+        // that is rather innocuous.
+        servers.shutDownAllServers();
+        waitForAll(servers, States.CONNECTING);
+        servers.restartAllServersAndClients(this);
+        waitForAll(servers, States.CONNECTED);
+
+        // 2. kill all followers
+        int leader = servers.findLeader();
+        Map<Long, Proposal> outstanding =  servers.mt[leader].main.quorumPeer.leader.outstandingProposals;
+        // increase the tick time to delay the leader going to looking
+        servers.mt[leader].main.quorumPeer.tickTime = LEADER_TIMEOUT_MS;
+        LOG.warn("LEADER " + leader);
+
+        for (int i = 0; i < SERVER_COUNT; i++) {
+            if (i != leader) {
+                servers.mt[i].shutdown();
+            }
+        }
+
+        // 3. start up the followers to form a new quorum
+        for (int i = 0; i < SERVER_COUNT; i++) {
+            if (i != leader) {
+                servers.mt[i].start();
+            }
+        }
+
+        // 4. wait one of the follower to be the new leader
+        for (int i = 0; i < SERVER_COUNT; i++) {
+            if (i != leader) {
+                // Recreate a client session since the previous session was not persisted.
+                servers.restartClient(i, this);
+                waitForOne(servers.zk[i], States.CONNECTED);
+            }
+        }
+
+        // 5. send a create request to old leader and make sure it's synced to disk,
+        //    which means it acked from itself
+        try {
+            servers.zk[leader].create("/zk" + leader, "zk".getBytes(), Ids.OPEN_ACL_UNSAFE,
+                CreateMode.PERSISTENT);
+            Assert.fail("create /zk" + leader + " should have failed");
+        } catch (KeeperException e) {
+        }
+
+        // just make sure that we actually did get it in process at the
+        // leader
+        Assert.assertEquals(1, outstanding.size());
+        Proposal p = outstanding.values().iterator().next();
+        Assert.assertEquals(OpCode.create, p.request.hdr.getType());
+
+        // make sure it has a chance to write it to disk
+        int sleepTime = 0;
+        Long longLeader = new Long(leader);
+        while (!p.ackSet.contains(longLeader)) {
+            if (sleepTime > 2000) {
+                Assert.fail("Transaction not synced to disk within 1 second " + p.ackSet
+                    + " expected " + leader);
+            }
+            Thread.sleep(100);
+            sleepTime += 100;
+        }
+
+        // 6. wait for the leader to quit due to not enough followers and come back up as a part of the new quorum
+        LOG.info("Waiting for leader " + leader + " to timeout followers");
+        sleepTime = 0;
+        Follower f = servers.mt[leader].main.quorumPeer.follower;
+        while (f == null || !f.isRunning()) {
+            if (sleepTime > LEADER_TIMEOUT_MS * 2) {
+                Assert.fail("Took too long for old leader to time out " + servers.mt[leader].main.quorumPeer.getPeerState());
+            }
+            Thread.sleep(100);
+            sleepTime += 100;
+            f = servers.mt[leader].main.quorumPeer.follower;
+        }
+
+        int newLeader = servers.findLeader();
+        // make sure a different leader was elected
+        Assert.assertTrue(leader != newLeader);
+
+        // 7. restart the previous leader to force it to replay the edits and possibly come up in a bad state
+        servers.mt[leader].shutdown();
+        servers.mt[leader].start();
+        waitForAll(servers, States.CONNECTED);
+
+        // 8. check the node exist in previous leader but not others
+        //    make sure everything is consistent
+        for (int i = 0; i < SERVER_COUNT; i++) {
+            Assert.assertNull("server " + i + " should not have /zk" + leader, servers.zk[i].exists("/zk" + leader, false));
+        }
+    }
 }
